from django.apps import AppConfig

from django_jinja import base


class DjangoJinjaAppConfig(AppConfig):
    name = "django_jinja"
    verbose_name = "Django Jinja"

    def ready(self):
        base.patch_django_for_autoescape()



import functools
from importlib import import_module

from django.utils.safestring import mark_safe
from django.core.exceptions import ImproperlyConfigured



def load_class(path):
    """
    Load class from path.
    """

    try:
        mod_name, klass_name = path.rsplit('.', 1)
        mod = import_module(mod_name)
    except AttributeError as e:
        raise ImproperlyConfigured(f'Error importing {mod_name}: "{e}"')

    try:
        klass = getattr(mod, klass_name)
    except AttributeError:
        raise ImproperlyConfigured(f'Module "{mod_name}" does not define a "{klass_name}" class')

    return klass


def safe(function):
    @functools.wraps(function)
    def _decorator(*args, **kwargs):
        return mark_safe(function(*args, **kwargs))
    return _decorator


def reraise(tp, value, tb=None):
    if value is None:
        value = tp()
    if value.__traceback__ is not tb:
        raise value.with_traceback(tb)
    raise value




"""Jinja2's i18n functionality is not exactly the same as Django's.
In particular, the tags names and their syntax are different:

  1. The Django ``trans`` tag is replaced by a _() global.
  2. The Django ``blocktrans`` tag is called ``trans``.

(1) isn't an issue, since the whole ``makemessages`` process is based on
converting the template tags to ``_()`` calls. However, (2) means that
those Jinja2 ``trans`` tags will not be picked up my Django's
``makemessage`` command.

There aren't any nice solutions here. While Jinja2's i18n extension does
come with extraction capabilities built in, the code behind ``makemessages``
unfortunately isn't extensible, so we can:

  * Duplicate the command + code behind it.
  * Offer a separate command for Jinja2 extraction.
  * Try to get Django to offer hooks into makemessages().
  * Monkey-patch.

We are currently doing that last thing. It turns out there we are lucky
for once: It's simply a matter of extending two regular expressions.
Credit for the approach goes to:
http://stackoverflow.com/questions/2090717/getting-translation-strings-for-jinja2-templates-integrated-with-django-1-x
"""

import re

from django.core.management.commands import makemessages
from django.template import engines
from django.template.base import BLOCK_TAG_START, BLOCK_TAG_END
from django.utils.translation import template as trans_real


strip_whitespace_right = re.compile(fr"({BLOCK_TAG_START}-?\s*(trans|pluralize).*?-{BLOCK_TAG_END})\s+", re.U)
strip_whitespace_left = re.compile(fr"\s+({BLOCK_TAG_START}-\s*(endtrans|pluralize).*?-?{BLOCK_TAG_END})", re.U)


def strip_whitespaces(src):
    src = strip_whitespace_left.sub(r'\1', src)
    src = strip_whitespace_right.sub(r'\1', src)
    return src


# this regex looks for {% trans %} blocks that don't have 'trimmed' or 'notrimmed' set.
# capturing {% endtrans %} ensures this doesn't affect DTL {% trans %} tags.
trans_block_re = re.compile(
    fr"({BLOCK_TAG_START}-?\s*trans)(?!\s+(?:no)?trimmed)"
    fr"(.*?{BLOCK_TAG_END}.*?{BLOCK_TAG_START}-?\s*?endtrans\s*?-?{BLOCK_TAG_END})",
    re.U | re.DOTALL
)


def apply_i18n_trimmed_policy(src, engine):
    # if env.policies["ext.i18n.trimmed"]: insert 'trimmed' flag on jinja {% trans %} blocks.
    i18n_trim_policy = engine.get("OPTIONS", {}).get("policies", {}).get("ext.i18n.trimmed")
    if not i18n_trim_policy:
        return src
    return trans_block_re.sub(r"\1 trimmed \2", src)


class Command(makemessages.Command):

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument('--jinja2-engine-name', default=None, dest='jinja_engine')

    def _get_default_jinja_template_engine(self):
        # dev's note: i would love to have this easy default: --jinja2-engine-name=jinja2
        # but as currently implemented, django-jinja's engine's name defaults to 'backend'.
        # see: https://docs.djangoproject.com/en/dev/ref/settings/#std:setting-TEMPLATES-NAME
        # for now, the default engine will be the first one exactly matching our backend's path.
        return [e for e in engines.templates.values() if e["BACKEND"] == "django_jinja.backend.Jinja2"][0]

    def handle(self, *args, **options):
        old_endblock_re = trans_real.endblock_re
        old_block_re = trans_real.block_re
        old_constant_re = trans_real.constant_re

        old_templatize = trans_real.templatize
        # Extend the regular expressions that are used to detect
        # translation blocks with an "OR jinja-syntax" clause.
        trans_real.endblock_re = re.compile(
            trans_real.endblock_re.pattern + '|' + r"""^-?\s*endtrans\s*-?$""")
        trans_real.block_re = re.compile(
            trans_real.block_re.pattern + '|' + r"""^-?\s*trans(?:\s+(?:no)?trimmed)?(?:\s+(?!'|")(?=.*?=.*?)|\s*-?$)""")
        trans_real.plural_re = re.compile(
            trans_real.plural_re.pattern + '|' + r"""^-?\s*pluralize(?:\s+.+|-?$)""")
        trans_real.constant_re = re.compile(r""".*?_\(((?:".*?(?<!\\)")|(?:'.*?(?<!\\)')).*?\)""")

        if options['jinja_engine']:
            jinja_engine = engines[options['jinja_engine']]
        else:
            jinja_engine = self._get_default_jinja_template_engine()

        def my_templatize(src, origin=None, **kwargs):
            new_src = strip_whitespaces(src)
            new_src = apply_i18n_trimmed_policy(new_src, jinja_engine)
            return old_templatize(new_src, origin, **kwargs)

        trans_real.templatize = my_templatize

        try:
            super().handle(*args, **options)
        finally:
            trans_real.endblock_re = old_endblock_re
            trans_real.block_re = old_block_re
            trans_real.templatize = old_templatize
            trans_real.constant_re = old_constant_re



DEFAULT_EXTENSIONS = [
    "jinja2.ext.do",
    "jinja2.ext.loopcontrols",
    "jinja2.ext.i18n",
    "django_jinja.builtins.extensions.DebugExtension",
    "django_jinja.builtins.extensions.CsrfExtension",
    "django_jinja.builtins.extensions.CacheExtension",
    "django_jinja.builtins.extensions.TimezoneExtension",
    "django_jinja.builtins.extensions.UrlsExtension",
    "django_jinja.builtins.extensions.StaticFilesExtension",
    "django_jinja.builtins.extensions.DjangoFiltersExtension",
]



from django.urls import re_path as url
from django_jinja import views

from .views import BasicTestView
from .views import I18nTestView, I18nTestViewDTL
from .views import StreamingTestView
from .views import CreateTestView, DeleteTestView, DetailTestView, UpdateTestView
from .views import ListTestView
from .views import ArchiveIndexTestView, YearArchiveTestView, MonthArchiveTestView, WeekArchiveTestView, DayArchiveTestView, TodayArchiveTestView, DateDetailTestView

urlpatterns = [
    url(r"^test1/$", BasicTestView.as_view(), name="test-1"),
    url(r"^test1/(?P<data>\d+)/$", BasicTestView.as_view(), name="test-1"),
    url(r"^test-i18n/$", I18nTestView.as_view(), name="i18n-test"),
    url(r"^test-i18n-dtl/$", I18nTestViewDTL.as_view(), name="i18n-dtl-test"),
    url(r"^test/404$", views.PageNotFound.as_view(), name="page-404"),
    url(r"^test/403$", views.PermissionDenied.as_view(), name="page-403"),
    url(r"^test/500$", views.ServerError.as_view(), name="page-500"),
    url(r"^test-streaming/$", StreamingTestView.as_view(), name='streaming-test'),

    url(r"^testmodel/$", ListTestView.as_view()),
    url(r"^testmodel/create$", CreateTestView.as_view()),
    url(r"^testmodel/(?P<pk>\d+)/delete$", DeleteTestView.as_view()),
    url(r"^testmodel/(?P<pk>\d+)/detail$", DetailTestView.as_view()),
    url(r"^testmodel/(?P<pk>\d+)/update$", UpdateTestView.as_view()),

    url(r"^testmodel/archive/$", ArchiveIndexTestView.as_view()),
    url(r"^testmodel/archive/(?P<year>\d{4})/$", YearArchiveTestView.as_view()),
    url(r"^testmodel/archive/(?P<year>\d{4})/week/(?P<week>\d+)/$", WeekArchiveTestView.as_view()),
    url(r"^testmodel/archive/(?P<year>\d{4})/(?P<month>[\w-]+)/$", MonthArchiveTestView.as_view()),
    url(r"^testmodel/archive/(?P<year>\d{4})/(?P<month>[\w-]+)/(?P<day>\d+)/$", DayArchiveTestView.as_view()),
    url(r"^testmodel/archive/today/$", TodayArchiveTestView.as_view()),
    url(r"^testmodel/archive/(?P<year>\d{4})/(?P<month>[\w-]+)/(?P<day>\d+)/(?P<pk>\d+)$", DateDetailTestView.as_view())
]



import datetime

from unittest import mock

from django.conf import global_settings
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.urls import reverse
from django.middleware import csrf
from django.shortcuts import render
from django.template import RequestContext
from django.template import engines
from django.template.loader import get_template
from django.test import TestCase
from django.test import override_settings
from django.test.client import RequestFactory
from django.utils import timezone
from django_jinja.backend import Jinja2
from django_jinja.base import get_match_extension
from django_jinja.base import match_template
from django_jinja.views.generic.base import Jinja2TemplateResponseMixin

from .forms import TestForm
from .models import TestModel
from .views import StreamingTestView


class RenderTemplatesTests(TestCase):
    def setUp(self):
        self.env = engines["jinja2"]
        self.factory = RequestFactory()

    def test_template_filters(self):
        filters_data = [
            ("{{ 'test-static.css'|static }}", {}, '/static/test-static.css'),
            ("{{ 'test-1'|reverseurl }}", {}, '/test1/'),
            ("{{ 'test-1'|reverseurl(data=2) }}", {}, '/test1/2/'),
            ("{{ num|floatformat }}", {'num': 34.23234}, '34.2'),
            ("{{ num|floatformat(3) }}", {'num': 34.23234}, '34.232'),
            ("{{ 'hola'|capfirst }}", {}, "Hola"),
            ("{{ 'hola mundo'|truncatechars(5) }}", {}, "hola…"),
            ("{{ 'hola mundo'|truncatechars_html(5) }}", {}, "hola…"),
            ("{{ 'hola mundo'|truncatewords(1) }}", {}, "hola …"),
            ("{{ 'hola mundo'|truncatewords_html(1) }}", {}, "hola …"),
            ("{{ 'hola mundo'|wordwrap(1) }}", {}, "hola\nmundo"),
            ("{{ 'hola mundo'|title }}", {}, "Hola Mundo"),
            ("{{ 'hola mundo'|slugify }}", {}, "hola-mundo"),
            ("{{ 'hello'|ljust(10) }}", {}, "hello     "),
            ("{{ 'hello'|rjust(10) }}", {}, "     hello"),
            ("{{ 'hello\nworld'|linebreaksbr }}", {}, "hello<br>world"),
            ("{{ '<div>hello</div>'|striptags }}", {}, "hello"),
            ("{{ list|join(',') }}", {'list':['a','b']}, 'a,b'),
            ("{{ 3|add(2) }}", {}, "5"),
            ("{{ now|date('n Y') }}", {"now": datetime.datetime(2012, 12, 20)}, "12 2012"),
            ("{{ url('test-1') }}", {}, '/test1/'),
            ("{{ foo }}", {}, "bar"),
        ]

        print()
        for template_str, kwargs, result in filters_data:
            print("- Testing: ", template_str, "with:", kwargs)
            template = self.env.from_string(template_str)
            _result = template.render(kwargs)
            self.assertEqual(_result, result)

    def test_string_interpolation(self):
        template = self.env.from_string("{{ 'Hello %s!' % name }}")
        self.assertEqual(template.render({"name": "foo"}), "Hello foo!")

        template = self.env.from_string("{{ _('Hello %s!').format(name) }}")
        self.assertEqual(template.render({"name": "foo"}), "Hello foo!")

    def test_urlresolve_exceptions(self):
        template = self.env.from_string("{{ url('adads') }}")
        template.render({})

    def test_custom_addons_01(self):
        template = self.env.from_string("{{ 'Hello'|replace('H','M') }}")
        result = template.render({})

        self.assertEqual(result, "Mello")

    def test_custom_addons_02(self):
        template = self.env.from_string("{% if m is one %}Foo{% endif %}")
        result = template.render({'m': 1})

        self.assertEqual(result, "Foo")

    def test_custom_addons_03(self):
        template = self.env.from_string("{{ myecho('foo') }}")
        result = template.render({})

        self.assertEqual(result, "foo")

    def test_render_with(self):
        template = self.env.from_string("{{ myrenderwith() }}")
        result = template.render({})
        self.assertEqual(result, "<strong>Foo</strong>")

    def test_django_context(self):
        """
        Test that Django context objects (which are stacks of dicts)
        can be passed directly to Jinja2 templates.
        """
        template = self.env.from_string("{{ greeting }}, {{ name }}")
        request = self.factory.get('/')
        ctx = RequestContext(request, {"greeting": "Hello", "name": "stranger"})
        with ctx.push(greeting="Hi"):
            with ctx.push(name="friend"):
                rendered1 = template.render(ctx)
        rendered2 = template.render(ctx)
        self.assertEqual(rendered1, "Hi, friend")
        self.assertEqual(rendered2, "Hello, stranger")

    def test_autoscape_with_form(self):
        form = TestForm()
        template = self.env.from_string("{{ form.as_p() }}")
        result = template.render({"form": form})

        self.assertIn('maxlength="2"', result)
        self.assertIn("<input ", result)

    def test_autoscape_with_form_field(self):
        form = TestForm()
        template = self.env.from_string("{{ form.name }}")
        result = template.render({"form": form})

        self.assertIn('maxlength="2"', result)
        self.assertIn("<input ", result)

    def test_autoscape_with_form_errors(self):
        form = TestForm({"name": "foo"})
        self.assertFalse(form.is_valid())

        template = self.env.from_string("{{ form.name.errors }}")
        result = template.render({"form": form})

        self.assertEqual(result,
                         ("""<ul class="errorlist"><li>Ensure this value """
                          """has at most 2 characters (it has 3).</li></ul>"""))

        template = self.env.from_string("{{ form.errors }}")
        result = template.render({"form": form})

        self.assertEqual(result,
                         ("""<ul class="errorlist"><li>name<ul class="errorlist">"""
                          """<li>Ensure this value has at most 2 characters (it """
                          """has 3).</li></ul></li></ul>"""))

    def test_autoescape_01(self):
        template = self.env.from_string("{{ foo|safe }}")
        result = template.render({'foo': '<h1>Hellp</h1>'})
        self.assertEqual(result, "<h1>Hellp</h1>")

    def test_autoescape_02(self):
        template = self.env.from_string("{{ foo }}")
        result = template.render({'foo': '<h1>Hellp</h1>'})
        self.assertEqual(result, "&lt;h1&gt;Hellp&lt;/h1&gt;")

    def test_autoescape_03(self):
        template = self.env.from_string("{{ foo|linebreaksbr }}")
        result = template.render({"foo": "<script>alert(1)</script>\nfoo"})
        self.assertEqual(result, "&lt;script&gt;alert(1)&lt;/script&gt;<br>foo")

    def test_debug_var_when_render_shortcut_is_used(self):
        prev_debug_value = settings.DEBUG
        settings.DEBUG = True

        request = self.factory.get("/")
        response = render(request, "test-debug-var.jinja")
        self.assertEqual(response.content, b"foobar")

        settings.DEBUG = prev_debug_value

    def test_debug_tag(self):
        """Test for {% debug %}"""
        tmpl = self.env.from_string('''Hello{% debug %}Bye''')
        out = tmpl.render()
        out = out.replace('&#39;', "'").replace('&lt;', '<').replace('&gt;', '>')
        #
        # Check that some of the built-in items exist in the debug output...
        #
        assert "'context'" in out
        assert "'cycler'" in out
        assert "'filters'" in out
        assert "'abs'" in out
        assert "'tests'" in out
        assert "'!='" in out

    def test_csrf_01(self):
        template_content = "{% csrf_token %}"

        with mock.patch("django.middleware.csrf.get_token",
                        return_value="123123") as m:

            request = self.factory.get('/customer/details')
            token = csrf.get_token(request)

            template = self.env.from_string(template_content)
            result = template.render({}, request)

            expected = f"<input type='hidden' name='csrfmiddlewaretoken' value='{token}' />"

        self.assertEqual(token, "123123")
        self.assertEqual(result, expected)

    def test_cache_01(self):
        template_content = "{% cache 200 'fooo' %}fóäo bar{% endcache %}"

        request = self.factory.get('/customer/details')
        context = RequestContext(request)

        template = self.env.from_string(template_content)
        result = template.render(context)

        self.assertEqual(result, "fóäo bar")

    def test_cache_none(self):
        template_content1 = "{% cache none 'barr' %}foo foo berry{% endcache %}"
        template_content2 = "{% cache none 'barr' %}not rendered{% endcache %}"

        request = self.factory.get('/forgotten/flavor')
        context = RequestContext(request)

        template1 = self.env.from_string(template_content1)
        result1 = template1.render(context)
        template2 = self.env.from_string(template_content2)
        result2 = template2.render(context)

        self.assertEqual(result1, "foo foo berry")
        self.assertEqual(result2, "foo foo berry")

    def test_404_page(self):
        response = self.client.get(reverse("page-404"))
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.content, b"404")
        response = self.client.post(reverse("page-404"))
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.content, b"404")
        response = self.client.put(reverse("page-404"))
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.content, b"404")
        response = self.client.delete(reverse("page-404"))
        self.assertEqual(response.status_code, 404)
        self.assertEqual(response.content, b"404")

    def test_403_page(self):
        response = self.client.get(reverse("page-403"))
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.content, b"403")

    def test_500_page(self):
        response = self.client.get(reverse("page-500"))
        self.assertEqual(response.status_code, 500)
        self.assertEqual(response.content, b"500")

    def test_get_default(self):
        Jinja2.get_default.cache_clear()
        self.assertEqual(Jinja2.get_default(), self.env)

    def test_get_default_multiple(self):
        setting = {
            "append": [
                {
                    "BACKEND": "django_jinja.backend.Jinja2",
                    "NAME": "jinja2dup",
                    "APP_DIRS": True,
                    "OPTIONS": {
                        "match_extension": ".jinjadup",
                    }
                }
            ]
        }

        with self.modify_settings(TEMPLATES=setting):
            with self.assertRaisesRegex(ImproperlyConfigured, r'Several Jinja2 backends'):
                Jinja2.get_default()

    def test_get_default_none(self):
        with self.settings(TEMPLATES=global_settings.TEMPLATES):
            with self.assertRaisesRegex(ImproperlyConfigured, r'No Jinja2 backend is configured'):
                Jinja2.get_default()


    def test_overwrite_default_app_dirname(self):
        setting = [
            {
                "BACKEND": "django_jinja.backend.Jinja2",
                "NAME": "jinja2",
                "APP_DIRS": True,
                "OPTIONS": {
                    "match_extension": None,
                    "match_regex": None,
                    "app_dirname": "jinja2",
                }
            }
        ]

        with override_settings(TEMPLATES=setting):
            template = get_template("hola_mundo.html")
            data = template.render({"name": "jinja2"})
            self.assertEqual(data, "hola mundo de jinja2")

    def test_context_manipulation(self):
        response = self.client.get(reverse("test-1"))
        self.assertEqual(response.context["name"], "Jinja2")
        self.assertTemplateUsed(response, 'hello_world.jinja')

    def test_streaming_response(self):
        template = "streaming_test.jinja"
        context = {"view": StreamingTestView, "name": "Streaming Jinja2"}
        response = self.client.get(reverse('streaming-test'))
        self.assertEqual(response.context["name"], context["name"])
        self.assertEqual(response.context["view"], context["view"])
        self.assertTemplateUsed(response, template)
        template = get_template(template)
        self.assertEqual(
            b''.join(response.streaming_content),
            template.render(context).encode()
        )


class BaseTests(TestCase):
    def setUp(self):
        self.env = engines["jinja2"]

    def test_match_template(self):
        self.assertTrue(
            match_template('admin/foo.html', regex=None, extension=None))
        self.assertFalse(
            match_template('admin/foo.html', regex=None, extension='.jinja'))
        self.assertTrue(
            match_template('admin/foo.html', regex=None, extension='.html'))
        self.assertTrue(
            match_template('admin/foo.html', regex=r'.*\.html', extension=None))
        self.assertFalse(
            match_template('admin/foo.html', regex=r"^(?!admin/.*)", extension=None))

    def test_get_match_extension(self):
        self.assertEqual(Jinja2.get_default().match_extension, get_match_extension())

    def test_get_match_extension_using(self):
        setting = {
            "append": [
                {
                    "BACKEND": "django_jinja.backend.Jinja2",
                    "NAME": "jinja2dup",
                    "APP_DIRS": True,
                    "OPTIONS": {
                        "match_extension": ".jinjadup",
                    }
                }
            ]
        }

        with self.modify_settings(TEMPLATES=setting):
            self.assertEqual(".jinja", get_match_extension(using='jinja2'))
            self.assertEqual(".jinjadup", get_match_extension(using="jinja2dup"))


class TemplateResponseTests(TestCase):
    class _BaseView:
        def get_template_names(self):
            return [
                'name1.html',
                'name2.html',
                'name3.html.jinja',
            ]

    def setUp(self):
        self.obj1 = TestModel.objects.create()

    def test_get_template_names(self):
        class _View(Jinja2TemplateResponseMixin, self._BaseView):
            pass

        view = _View()
        self.assertEqual(
            ['name1.html.jinja', 'name2.html.jinja', 'name3.html.jinja'],
            view.get_template_names()
        )

    def test_get_template_names_classext(self):
        class _View(Jinja2TemplateResponseMixin, self._BaseView):
            jinja2_template_extension = '.foo'

        view = _View()
        self.assertEqual(
            ['name1.html.foo', 'name2.html.foo', 'name3.html.jinja.foo'],
            view.get_template_names()
        )

    def test_get_template_names_using(self):
        class _View(Jinja2TemplateResponseMixin, self._BaseView):
            template_engine = 'jinja2dup'

        setting = {
            "append": [
                {
                    "BACKEND": "django_jinja.backend.Jinja2",
                    "NAME": "jinja2dup",
                    "APP_DIRS": True,
                    "OPTIONS": {
                        "match_extension": ".jinjadup",
                    }
                }
            ]
        }

        with self.modify_settings(TEMPLATES=setting):
            view = _View()
            self.assertEqual(
                ['name1.html.jinjadup', 'name2.html.jinjadup', 'name3.html.jinja.jinjadup'],
                view.get_template_names()
            )

class GenericViewTests(TestCase):
    def setUp(self):
        self.obj1 = TestModel.objects.create(date=timezone.now())
        self.obj2 = TestModel.objects.create(date=timezone.now())

    def test_detailview(self):
        self.assertContains(
            self.client.get(f'/testmodel/{self.obj1.pk}/detail'),
            'DetailView Test Template',
            status_code=200
        )

    def test_createview(self):
        self.assertContains(
            self.client.get('/testmodel/create'),
            'CreateView Test Template',
            status_code=200
        )

    def test_deleteview(self):
        self.assertContains(
            self.client.get(f'/testmodel/{self.obj1.pk}/delete'),
            'DeleteView Test Template',
            status_code=200
        )

    def test_updateview(self):
        self.assertContains(
            self.client.get(f'/testmodel/{self.obj1.pk}/update'),
            'UpdateView Test Template',
            status_code=200
        )

    def test_listview(self):
        self.assertContains(
            self.client.get('/testmodel/'),
            'ListView Test Template',
            status_code=200
        )

    def test_archiveindexview(self):
        self.assertContains(
            self.client.get('/testmodel/archive/'),
            'ArchiveIndexView Test Template',
            status_code=200
        )

    def test_yeararchiveview(self):
        self.assertContains(
            self.client.get(f'/testmodel/archive/{self.obj1.date:%Y}/'),
            'YearArchiveView Test Template',
            status_code=200
        )

    def test_montharchiveview(self):
        self.assertContains(
            self.client.get(f'/testmodel/archive/{self.obj1.date:%Y/%b}/'),
            'MonthArchiveView Test Template',
            status_code=200
        )

    def test_weekarchiveview(self):
        self.assertContains(
            self.client.get(f'/testmodel/archive/{self.obj1.date:%Y/week/%U}/'),
            'WeekArchiveView Test Template',
            status_code=200
        )

    def test_dayarchiveview(self):
        self.assertContains(
            self.client.get(f'/testmodel/archive/{self.obj1.date:%Y/%b/%d}/'),
            'DayArchiveView Test Template',
            status_code=200
        )

    def test_todayarchiveview(self):
        self.assertContains(
            self.client.get('/testmodel/archive/today/'),
            'TodayArchiveView Test Template',
            status_code=200
        )

    def test_datedetailview(self):
        self.assertContains(
            self.client.get(f'/testmodel/archive/{self.obj1.date:%Y/%b/%d}/{self.obj1.pk}'),
            'DateDetailView Test Template',
            status_code=200
        )

    # ==== Special imports ====
    def test_import_archiveindexview(self):
        from django_jinja.views.generic import ArchiveIndexView

    def test_import_yeararchiveview(self):
        from django_jinja.views.generic import YearArchiveView

    def test_import_montharchiveview(self):
        from django_jinja.views.generic import MonthArchiveView

    def test_import_dayarchiveview(self):
        from django_jinja.views.generic import DayArchiveView

    def test_import_weekarchiveview(self):
        from django_jinja.views.generic import WeekArchiveView

    def test_import_todayarchiveview(self):
        from django_jinja.views.generic import TodayArchiveView

    def test_import_datedetailview(self):
        from django_jinja.views.generic import DateDetailView

    def test_import_detailview(self):
        from django_jinja.views.generic import DetailView

    def test_import_createview(self):
        from django_jinja.views.generic import CreateView

    def test_import_updateview(self):
        from django_jinja.views.generic import UpdateView

    def test_import_deleteview(self):
        from django_jinja.views.generic import DeleteView

    def test_import_listview(self):
        from django_jinja.views.generic import ListView



import functools

from django.template.loader import render_to_string
from django.utils.safestring import mark_safe


# Global register dict for third party
# template functions, filters and extensions.
_local_env = {
    "globals": {},
    "tests": {},
    "filters": {},
    "extensions": set(),
}


def _update_env(env):
    """
    Given a jinja environment, update it with third party
    collected environment extensions.
    """

    env.globals.update(_local_env["globals"])
    env.tests.update(_local_env["tests"])
    env.filters.update(_local_env["filters"])

    for extension in _local_env["extensions"]:
        env.add_extension(extension)


def _attach_function(attr, func, name=None):
    if name is None:
        name = func.__name__

    global _local_env
    _local_env[attr][name] = func
    return func


def _register_function(attr, name=None, fn=None):
    if name is None and fn is None:
        def dec(func):
            return _attach_function(attr, func)
        return dec

    elif name is not None and fn is None:
        if callable(name):
            return _attach_function(attr, name)
        else:
            def dec(func):
                return _register_function(attr, name, func)
            return dec

    elif name is not None and fn is not None:
        return _attach_function(attr, fn, name)

    raise RuntimeError("Invalid parameters")


def extension(extension):
    global _local_env
    _local_env["extensions"].add(extension)
    return extension


def global_function(*args, **kwargs):
    return _register_function("globals", *args, **kwargs)


def test(*args, **kwargs):
    return _register_function("tests", *args, **kwargs)


def filter(*args, **kwargs):
    return _register_function("filters", *args, **kwargs)


def render_with(template, fn=None):
    """
    Makes simple function works like
    django's default inclusion_tag: render
    specified template with context returned
    by decorated function.
    """

    if fn is None:
        return functools.partial(render_with, template)

    @functools.wraps(fn)
    def _wrapper(*args, **kwargs):
        data = render_to_string(template, fn(*args, **kwargs))
        return mark_safe(data)

    return _wrapper



import re
import os.path as path
from importlib import import_module

from django.template.context import BaseContext


def dict_from_context(context):
    """
    Converts context to native python dict.
    """

    if isinstance(context, BaseContext):
        new_dict = {}
        for i in reversed(list(context)):
            new_dict.update(dict_from_context(i))
        return new_dict

    return dict(context)


def _iter_templatetags_modules_list():
    """
    Get list of modules that contains templatetags
    submodule.
    """
    from django.apps import apps
    all_modules = [x.name for x in apps.get_app_configs()]

    for app_path in all_modules:
        try:
            mod = import_module(app_path + ".templatetags")
            # Empty folders can lead to unexpected behavior with Python 3.
            # We make sure to have the `__file__` attribute.
            if getattr(mod, '__file__', None) is not None:
                yield (app_path, path.dirname(mod.__file__))
        except ImportError:
            pass


def patch_django_for_autoescape():
    """
    Patch django modules for make them compatible with
    jinja autoescape implementation.
    """
    from django.utils import safestring
    from django.forms.boundfield import BoundField
    from django.forms.utils import ErrorList
    from django.forms.utils import ErrorDict

    if hasattr(safestring, "SafeText"):
        if not hasattr(safestring.SafeText, "__html__"):
            safestring.SafeText.__html__ = lambda self: str(self)

    if hasattr(safestring, "SafeString"):
        if not hasattr(safestring.SafeString, "__html__"):
            safestring.SafeString.__html__ = lambda self: str(self)

    if hasattr(safestring, "SafeUnicode"):
        if not hasattr(safestring.SafeUnicode, "__html__"):
            safestring.SafeUnicode.__html__ = lambda self: str(self)

    if hasattr(safestring, "SafeBytes"):
        if not hasattr(safestring.SafeBytes, "__html__"):
            safestring.SafeBytes.__html__ = lambda self: str(self)

    if not hasattr(BoundField, "__html__"):
        BoundField.__html__ = lambda self: str(self)

    if not hasattr(ErrorList, "__html__"):
        ErrorList.__html__ = lambda self: str(self)

    if not hasattr(ErrorDict, "__html__"):
        ErrorDict.__html__ = lambda self: str(self)


def get_match_extension(using=None):
    """
    Gets the extension that the template loader will match for
    django-jinja. This returns Jinja2.match_extension.

    The "using" parameter selects with Jinja2 backend to use if
    you have multiple ones configured in settings.TEMPLATES.
    If it is None and only one Jinja2 backend is defined then it
    will use that, otherwise an ImproperlyConfigured exception
    is thrown.
    """
    from .backend import Jinja2
    from django.template import engines

    if using is None:
        engine = Jinja2.get_default()
    else:
        engine = engines[using]

    return engine.match_extension


def match_template(template_name, extension, regex):
    if extension:
        matches_extension = template_name.endswith(extension)
        if regex:
            return matches_extension and re.match(regex, template_name)
        else:
            return matches_extension
    elif regex:
        return re.match(regex, template_name)
    else:
        return True



import logging
import pprint

from django.conf import settings
from django.contrib.staticfiles.storage import staticfiles_storage
from django.core.cache import cache
from django.core.cache.utils import make_template_fragment_key
from django.urls import NoReverseMatch
from django.urls import reverse
from django.utils.encoding import force_str
from jinja2.nodes import ContextReference
from jinja2 import TemplateSyntaxError
from jinja2 import pass_context
from jinja2 import nodes
from jinja2.ext import Extension
from markupsafe import Markup


JINJA2_MUTE_URLRESOLVE_EXCEPTIONS = getattr(settings, "JINJA2_MUTE_URLRESOLVE_EXCEPTIONS", False)
logger = logging.getLogger(__name__)


class CsrfExtension(Extension):
    tags = {'csrf_token'}

    def __init__(self, environment):
        self.environment = environment

    def parse(self, parser):
        lineno = parser.stream.expect('name:csrf_token').lineno
        call = self.call_method(
            '_render',
            [nodes.Name('csrf_token', 'load', lineno=lineno)],
            lineno=lineno
        )
        return nodes.Output([nodes.MarkSafe(call)])

    def _render(self, csrf_token):
        if csrf_token:
            if csrf_token == 'NOTPROVIDED':
                return Markup("")

            return Markup(f"<input type='hidden' name='csrfmiddlewaretoken' value='{csrf_token}' />")

        if settings.DEBUG:
            import warnings
            warnings.warn("A {% csrf_token %} was used in a template, but the context"
                          "did not provide the value.  This is usually caused by not "
                          "using RequestContext.")
        return ''


class CacheExtension(Extension):
    """
    Exactly like Django's own tag, but supports full Jinja2
    expressiveness for all arguments.

        {% cache gettimeout()*2 "foo"+options.cachename  %}
            ...
        {% endcache %}

    General Syntax:

        {% cache [expire_time] [fragment_name] [var1] [var2] .. %}
            .. some expensive processing ..
        {% endcache %}

    Available by default (does not need to be loaded).

    Partly based on the ``FragmentCacheExtension`` from the Jinja2 docs.
    """

    tags = {'cache'}

    def parse(self, parser):
        lineno = next(parser.stream).lineno

        expire_time = parser.parse_expression()
        fragment_name = parser.parse_expression()
        vary_on = []

        while not parser.stream.current.test('block_end'):
            vary_on.append(parser.parse_expression())

        body = parser.parse_statements(['name:endcache'], drop_needle=True)

        return nodes.CallBlock(
            self.call_method('_cache_support',
                             [expire_time, fragment_name,
                              nodes.List(vary_on), nodes.Const(lineno)]),
            [], [], body).set_lineno(lineno)

    def _cache_support(self, expire_time, fragm_name, vary_on, lineno, caller):
        try:
            if expire_time is not None:
                expire_time = int(expire_time)
        except (ValueError, TypeError):
            raise TemplateSyntaxError(
                f'"{list(self.tags)[0]}" tag got a non-integer timeout value: {expire_time!r}',
                lineno,
            )

        cache_key = make_template_fragment_key(fragm_name, vary_on)

        value = cache.get(cache_key)
        if value is None:
            value = caller()
            cache.set(cache_key, force_str(value), expire_time)
        else:
            value = force_str(value)

        return value


class DebugExtension(Extension):
    """
    A ``{% debug %}`` tag that dumps the available variables, filters and tests.
    Typical usage like this:

    .. codeblock:: html+jinja
        <pre>{% debug %}</pre>

    produces output like this:

    ::
        {'context': {'_': <function _gettext_alias at 0x7f9ceabde488>,
                 'csrf_token': <SimpleLazyObject: 'lfPE7al...q3bykS4txKfb3'>,
                 'cycler': <class 'jinja2.utils.Cycler'>,
                 ...
                 'view': <polls.views_auth.Login object at 0x7f9cea2cbe48>},
        'filters': ['abs', 'add', 'addslashes', 'attr', 'batch', 'bootstrap',
                 'bootstrap_classes', 'bootstrap_horizontal',
                 'bootstrap_inline', ... 'yesno'],
        'tests': ['callable', 'checkbox_field', 'defined', 'divisibleby',
               'escaped', 'even', 'iterable', 'lower', 'mapping',
               'multiple_checkbox_field', ... 'string', 'undefined', 'upper']}

    """
    tags = {'debug'}

    def __init__(self, environment):
        super().__init__(environment)

    def parse(self, parser):
        lineno = parser.stream.expect('name:debug').lineno
        context = ContextReference()
        call = self.call_method('_render', [context], lineno=lineno)
        return nodes.Output([nodes.MarkSafe(call)])

    def _render(self, context):
        result = {
            'filters': sorted(self.environment.filters.keys()),
            'tests': sorted(self.environment.tests.keys()),
            'context': context.get_all()
        }
        #
        # We set the depth since the intent is basically to show the top few
        # names. TODO: provide user control over this?
        #
        text = pprint.pformat(result, depth=3, compact=True)
        return Markup.escape(text)


class StaticFilesExtension(Extension):
    def __init__(self, environment):
        super().__init__(environment)
        environment.globals["static"] = self._static

    def _static(self, path):
        return staticfiles_storage.url(path)


class UrlsExtension(Extension):
    def __init__(self, environment):
        super().__init__(environment)
        environment.globals["url"] = self._url_reverse

    @pass_context
    def _url_reverse(self, context, name, *args, **kwargs):
        try:
            current_app = context["request"].current_app
        except AttributeError:
            try:
                current_app = context["request"].resolver_match.namespace
            except AttributeError:
                current_app = None
        except KeyError:
            current_app = None
        try:
            return reverse(name, args=args, kwargs=kwargs, current_app=current_app)
        except NoReverseMatch as exc:
            logger.error('Error: %s', exc)
            if not JINJA2_MUTE_URLRESOLVE_EXCEPTIONS:
                raise
            return ''
        return reverse(name, args=args, kwargs=kwargs)


from . import filters

class TimezoneExtension(Extension):
    def __init__(self, environment):
        super().__init__(environment)
        environment.globals["utc"] = filters.utc
        environment.globals["timezone"] = filters.timezone
        environment.globals["localtime"] = filters.localtime


class DjangoFiltersExtension(Extension):
    def __init__(self, environment):
        super().__init__(environment)
        environment.filters["static"] = filters.static
        environment.filters["reverseurl"] = filters.reverse
        environment.filters["addslashes"] = filters.addslashes
        environment.filters["capfirst"] = filters.capfirst
        environment.filters["escapejs"] = filters.escapejs_filter
        environment.filters["floatformat"] = filters.floatformat
        environment.filters["iriencode"] = filters.iriencode
        environment.filters["linenumbers"] = filters.linenumbers
        environment.filters["make_list"] = filters.make_list
        environment.filters["slugify"] = filters.slugify
        environment.filters["stringformat"] = filters.stringformat
        environment.filters["truncatechars"] = filters.truncatechars
        environment.filters["truncatechars_html"] = filters.truncatechars_html
        environment.filters["truncatewords"] = filters.truncatewords
        environment.filters["truncatewords_html"] = filters.truncatewords_html
        environment.filters["urlizetrunc"] = filters.urlizetrunc
        environment.filters["ljust"] = filters.ljust
        environment.filters["rjust"] = filters.rjust
        environment.filters["cut"] = filters.cut
        environment.filters["linebreaksbr"] = filters.linebreaksbr
        environment.filters["linebreaks"] = filters.linebreaks_filter
        environment.filters["striptags"] = filters.striptags
        environment.filters["add"] = filters.add
        environment.filters["date"] = filters.date
        environment.filters["time"] = filters.time
        environment.filters["timesince"] = filters.timesince_filter
        environment.filters["timeuntil"] = filters.timeuntil_filter
        environment.filters["default_if_none"] = filters.default_if_none
        environment.filters["divisibleby"] = filters.divisibleby
        environment.filters["yesno"] = filters.yesno
        environment.filters["pluralize"] = filters.pluralize
        environment.filters["localtime"] = filters.localtime
        environment.filters["utc"] = filters.utc
        environment.filters["timezone"] = filters.timezone
        environment.filters["json_script"] = filters.json_script


class DjangoExtraFiltersExtension(Extension):
    def __init__(self, environment):
        super().__init__(environment)
        environment.filters["title"] = filters.title
        environment.filters["upper"] = filters.upper
        environment.filters["lower"] = filters.lower
        environment.filters["urlencode"] = filters.urlencode
        environment.filters["urlize"] = filters.urlize
        environment.filters["wordcount"] = filters.wordcount
        environment.filters["wordwrap"] = filters.wordwrap
        environment.filters["center"] = filters.center
        environment.filters["join"] = filters.join
        environment.filters["length"] = filters.length
        environment.filters["random"] = filters.random
        environment.filters["default"] = filters.default
        environment.filters["filesizeformat"] = filters.filesizeformat
        environment.filters["pprint"] = filters.pprint



"""
Since django 1.8.x, django comes with native multiple template engine support.
It also comes with jinja2 backend, but it is slightly unflexible, and it does
not support by default all django filters and related stuff.

This is an implementation of django backend inteface for use
django_jinja easy with django 1.8.
"""

import sys
import os
import os.path as path
import functools
from importlib import import_module

import jinja2
from django.conf import settings
from django.core import signals
from django.core.exceptions import ImproperlyConfigured
from django.dispatch import receiver
from django.middleware import csrf
from django.template import TemplateDoesNotExist
from django.template import TemplateSyntaxError
from django.template.backends.base import BaseEngine
from django.template.context import BaseContext
from django.utils.encoding import smart_str
from django.utils.functional import SimpleLazyObject
from django.utils.functional import cached_property
from django.utils.module_loading import import_string
from django.utils.safestring import mark_safe

from . import base
from . import builtins
from . import library
from . import utils


class Origin:
    """
    A container to hold debug information as described in the template API
    documentation.
    """
    def __init__(self, name, template_name):
        self.name = name
        self.template_name = template_name


class Template:
    def __init__(self, template, backend):
        self.template = template
        self.backend = backend

        self.name = template.name
        self.origin = Origin(
            name=template.filename, template_name=template.name
        )

    def render(self, context=None, request=None):
        return mark_safe(self._process_template(self.template.render, context, request))

    def stream(self, context=None, request=None):
        return self._process_template(self.template.stream, context, request)

    def _process_template(self, handler, context=None, request=None):
        if context is None:
            context = {}

        context = base.dict_from_context(context)

        if request is not None:
            def _get_val():
                token = csrf.get_token(request)
                if token is None:
                    return 'NOTPROVIDED'
                else:
                    return smart_str(token)

            context["request"] = request
            context["csrf_token"] = SimpleLazyObject(_get_val)

            # Support for django context processors
            for processor in self.backend.context_processors:
                context.update(processor(request))

        if self.backend._tmpl_debug:
            from django.test import signals

            # Define a "django" like context for emitatet the multi
            # layered context object. This is mainly for apps like
            # django-debug-toolbar that are very coupled to django's
            # internal implementation of context.

            if not isinstance(context, BaseContext):
                class CompatibilityContext(dict):
                    @property
                    def dicts(self):
                        return [self]

                context = CompatibilityContext(context)

            signals.template_rendered.send(sender=self,
                                           template=self,
                                           context=context)

        return handler(context)


class Jinja2(BaseEngine):
    app_dirname = "templates"

    @staticmethod
    @functools.lru_cache()
    def get_default():
        """
        When only one django-jinja backend is configured, returns it.
        Raises ImproperlyConfigured otherwise.

        This is required for finding the match extension where the
        developer does not specify a template_engine on a
        TemplateResponseMixin subclass.
        """
        from django.template import engines

        jinja_engines = [engine for engine in engines.all()
                         if isinstance(engine, Jinja2)]
        if len(jinja_engines) == 1:
            # Unwrap the Jinja2 engine instance.
            return jinja_engines[0]
        elif len(jinja_engines) == 0:
            raise ImproperlyConfigured(
                "No Jinja2 backend is configured.")
        else:
            raise ImproperlyConfigured(
                "Several Jinja2 backends are configured. "
                "You must select one explicitly.")

    def __init__(self, params):
        params = params.copy()
        options = params.pop("OPTIONS", {}).copy()

        self.app_dirname = options.pop("app_dirname", "templates")
        super().__init__(params)

        newstyle_gettext = options.pop("newstyle_gettext", True)
        context_processors = options.pop("context_processors", [])
        match_extension = options.pop("match_extension", ".jinja")
        match_regex = options.pop("match_regex", None)
        environment_clspath = options.pop("environment", "jinja2.Environment")
        extra_filters = options.pop("filters", {})
        extra_tests = options.pop("tests", {})
        extra_globals = options.pop("globals", {})
        extra_constants = options.pop("constants", {})
        translation_engine = options.pop("translation_engine", "django.utils.translation")
        policies = options.pop("policies", {})

        tmpl_debug = options.pop("debug", settings.DEBUG)
        bytecode_cache = options.pop("bytecode_cache", {})
        bytecode_cache.setdefault("name", "default")
        bytecode_cache.setdefault("enabled", False)
        bytecode_cache.setdefault("backend", "django_jinja.cache.BytecodeCache")

        undefined = options.pop("undefined", None)
        if undefined is not None:
            if isinstance(undefined, str):
                options["undefined"] = utils.load_class(undefined)
            else:
                options["undefined"] = undefined

        if settings.DEBUG:
            options.setdefault("undefined", jinja2.DebugUndefined)
        else:
            options.setdefault("undefined", jinja2.Undefined)

        environment_cls = import_string(environment_clspath)

        if isinstance(options.get("loader"), str):
            # Allow to specify a loader as string
            loader_cls = import_string(options.pop("loader"))
        else:
            # Backward compatible default
            loader_cls = jinja2.FileSystemLoader

        options.setdefault("loader", loader_cls(self.template_dirs))
        options.setdefault("extensions", builtins.DEFAULT_EXTENSIONS)
        options.setdefault("auto_reload", settings.DEBUG)
        options.setdefault("autoescape", True)

        self.env = environment_cls(**options)

        # Initialize i18n support
        if settings.USE_I18N:
            translation = import_module(translation_engine)
            self.env.install_gettext_translations(translation, newstyle=newstyle_gettext)
        else:
            self.env.install_null_translations(newstyle=newstyle_gettext)

        self._context_processors = context_processors
        self._match_regex = match_regex
        self._match_extension = match_extension
        self._tmpl_debug = tmpl_debug
        self._bytecode_cache = bytecode_cache

        self._initialize_builtins(filters=extra_filters,
                                  tests=extra_tests,
                                  globals=extra_globals,
                                  constants=extra_constants)
        self._initialize_policies(policies)

        self._initialize_thirdparty()
        self._initialize_bytecode_cache()

    def _initialize_bytecode_cache(self):
        if self._bytecode_cache["enabled"]:
            cls = utils.load_class(self._bytecode_cache["backend"])
            self.env.bytecode_cache = cls(self._bytecode_cache["name"])

    def _initialize_thirdparty(self):
        """
        Iterate over all available apps in searching and preloading
        available template filters or functions for jinja2.
        """
        for app_path, mod_path in base._iter_templatetags_modules_list():
            if not path.isdir(mod_path):
                continue

            for filename in filter(lambda x: x.endswith(".py") or x.endswith(".pyc"), os.listdir(mod_path)):
                # Exclude __init__.py files
                if filename == "__init__.py" or filename == "__init__.pyc":
                    continue

                file_mod_path = f"{app_path}.templatetags.{filename.rsplit('.', 1)[0]}"
                try:
                    import_module(file_mod_path)
                except ImportError:
                    pass

            library._update_env(self.env)

    def _initialize_builtins(self, filters=None, tests=None, globals=None, constants=None):
        def insert(data, name, value):
            if isinstance(value, str):
                data[name] = import_string(value)
            else:
                data[name] = value

        if filters:
            for name, value in filters.items():
                insert(self.env.filters, name, value)

        if tests:
            for name, value in tests.items():
                insert(self.env.tests, name, value)

        if globals:
            for name, value in globals.items():
                insert(self.env.globals, name, value)

        if constants:
            for name, value in constants.items():
                self.env.globals[name] = value

    def _initialize_policies(self, policies):
        # Set policies like those in jinja2.defaults.DEFAULT_POLICIES
        for name, value in policies.items():
            self.env.policies[name] = value

    @cached_property
    def context_processors(self):
        return tuple(import_string(path) for path in self._context_processors)

    @property
    def match_extension(self):
        return self._match_extension

    def from_string(self, template_code):
        return Template(self.env.from_string(template_code), self)

    def match_template(self, template_name):
        return base.match_template(template_name,
                                   self._match_extension,
                                   self._match_regex)

    def get_template(self, template_name):
        if not self.match_template(template_name):
            message = f"Template {template_name} does not exists"
            raise TemplateDoesNotExist(message)

        try:
            return Template(self.env.get_template(template_name), self)
        except jinja2.TemplateNotFound as exc:
            # Unlike django's template engine, jinja2 doesn't like windows-style path separators.
            # But because django does, its docs encourage the usage of os.path.join().
            # Rather than insisting that our users switch to posixpath.join(), this try block
            # will attempt to retrieve the template path again with forward slashes on windows:
            if os.name == 'nt' and '\\' in template_name:
                try:
                    return self.get_template(template_name.replace("\\", "/"))
                except jinja2.TemplateNotFound:
                    pass

            exc = TemplateDoesNotExist(exc.name, backend=self)

            utils.reraise(
                TemplateDoesNotExist,
                exc,
                sys.exc_info()[2],
            )
        except jinja2.TemplateSyntaxError as exc:
            new = TemplateSyntaxError(exc.args)
            new.template_debug = get_exception_info(exc)
            utils.reraise(TemplateSyntaxError, new, sys.exc_info()[2])


@receiver(signals.setting_changed)
def _setting_changed(sender, setting, *args, **kwargs):
    """ Reset the Jinja2.get_default() cached when TEMPLATES changes. """
    if setting == "TEMPLATES":
        Jinja2.get_default.cache_clear()


def get_exception_info(exception):
    """
    Formats exception information for display on the debug page using the
    structure described in the template API documentation.
    """
    context_lines = 10
    lineno = exception.lineno
    if exception.source is None:
        if os.path.exists(exception.filename):
            with open(exception.filename) as f:
                source = f.read()
    else:
        source = exception.source
    lines = list(enumerate(source.strip().split("\n"), start=1))
    during = lines[lineno - 1][1]
    total = len(lines)
    top = max(0, lineno - context_lines - 1)
    bottom = min(total, lineno + context_lines)

    return {
        'name': exception.filename,
        'message': exception.message,
        'source_lines': lines[top:bottom],
        'line': lineno,
        'before': '',
        'during': during,
        'after': '',
        'total': total,
        'top': top,
        'bottom': bottom,
    }



from ...base import get_match_extension


class Jinja2TemplateResponseMixin:
    jinja2_template_extension = None

    def get_template_names(self):
        """
        Return a list of template names to be used for the request.
        This calls the super class's get_template_names and appends
        the Jinja2 match extension is suffixed to the returned values.

        If you specify jinja2_template_extension then that value will
        be used. Otherwise it tries to detect the extension based on
        values in settings.

        If you would like to not have it append an extension, set
        jinja2_template_extension to '' (empty string).
        """
        vals = super().get_template_names()

        ext = self.jinja2_template_extension
        if ext is None:
            ext = get_match_extension(using=getattr(self, 'template_engine', None))

        # Exit early if the user has specified an empty match extension
        if not ext:
            return vals

        names = []
        for val in vals:
            if not val.endswith(ext):
                val += ext
            names.append(val)

        return names



from django import http
from django.template import loader
from django.views import View

from ..base import get_match_extension


class GenericView(View):
    response_cls = http.HttpResponse
    content_type = "text/html"
    tmpl_name = None

    def get_context_data(self):
        return {"view": self}

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        template_name = callable(self.tmpl_name) and self.tmpl_name() or self.tmpl_name
        output = loader.render_to_string(template_name, context, request=request)
        return self.response_cls(output, content_type=self.content_type)


class ErrorView(GenericView):
    def head(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def options(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)


class PageNotFound(ErrorView):
    response_cls = http.HttpResponseNotFound

    def tmpl_name(self):
        return "404" + (get_match_extension() or ".jinja")


class PermissionDenied(ErrorView):
    response_cls = http.HttpResponseForbidden

    def tmpl_name(self):
        return "403" + (get_match_extension() or ".jinja")


class BadRequest(ErrorView):
    response_cls = http.HttpResponseBadRequest

    def tmpl_name(self):
        return "400" + (get_match_extension() or ".jinja")


class ServerError(ErrorView):
    response_cls = http.HttpResponseServerError

    def tmpl_name(self):
        return "500" + (get_match_extension() or ".jinja")



from django.utils.encoding import force_str
from django.urls import reverse as django_reverse
from django.contrib.staticfiles.storage import staticfiles_storage


def reverse(value, *args, **kwargs):
    """
    Shortcut filter for reverse url on templates. Is a alternative to
    django {% url %} tag, but more simple.

    Usage example:
        {{ 'web:timeline'|reverse(userid=2) }}

    This is a equivalent to django:
        {% url 'web:timeline' userid=2 %}

    """
    return django_reverse(value, args=args, kwargs=kwargs)

def static(path):
    return staticfiles_storage.url(path)


from django.template.defaultfilters import addslashes
from django.template.defaultfilters import capfirst
from django.utils.html import escapejs as escapejs_filter
# from django.utils.html import fix_ampersands as fix_ampersands_filter
from django.template.defaultfilters import floatformat
from django.template.defaultfilters import iriencode
from django.template.defaultfilters import linenumbers
from django.template.defaultfilters import make_list
from django.template.defaultfilters import stringformat
from django.template.defaultfilters import title
from django.template.defaultfilters import truncatechars
from django.template.defaultfilters import truncatechars_html
from django.template.defaultfilters import truncatewords
from django.template.defaultfilters import truncatewords_html
from django.template.defaultfilters import upper
from django.template.defaultfilters import lower
from django.template.defaultfilters import urlencode
from django.template.defaultfilters import urlize
from django.template.defaultfilters import urlizetrunc
from django.template.defaultfilters import wordcount
from django.template.defaultfilters import wordwrap
from django.template.defaultfilters import ljust
from django.template.defaultfilters import rjust
from django.template.defaultfilters import center
from django.template.defaultfilters import cut
from django.template.defaultfilters import linebreaks_filter
from django.template.defaultfilters import linebreaksbr
from django.template.defaultfilters import striptags
from django.template.defaultfilters import join
from django.template.defaultfilters import length
from django.template.defaultfilters import random
from django.template.defaultfilters import add
from django.template.defaultfilters import date
from django.template.defaultfilters import time
from django.template.defaultfilters import timesince_filter
from django.template.defaultfilters import timeuntil_filter
from django.template.defaultfilters import default
from django.template.defaultfilters import default_if_none
from django.template.defaultfilters import divisibleby
from django.template.defaultfilters import yesno
from django.template.defaultfilters import filesizeformat
from django.template.defaultfilters import pprint
from django.template.defaultfilters import pluralize
from django.template.defaultfilters import json_script

from django.utils.text import slugify as djslugify

def slugify(value):
    return djslugify(force_str(value))

from functools import partial
linebreaksbr = partial(linebreaksbr, autoescape=True)

# TZ

from django.templatetags.tz import do_timezone as timezone
from django.templatetags.tz import localtime
from django.templatetags.tz import utc



from django.views.generic import View
from django.http import HttpResponse, StreamingHttpResponse
from django.shortcuts import render
from django.template import loader
from django.template.loader import render_to_string

from django_jinja.views.generic.detail import DetailView
from django_jinja.views.generic.edit import CreateView, DeleteView, UpdateView
from django_jinja.views.generic.list import ListView
from django_jinja.views.generic.dates import ArchiveIndexView, YearArchiveView, MonthArchiveView, WeekArchiveView, DayArchiveView, TodayArchiveView, DateDetailView

from .models import TestModel


class BasicTestView(View):
    def get(self, request, data=None):
        data = render_to_string("hello_world.jinja", {"name": "Jinja2"},
                                request=request)
        return HttpResponse(data)

class I18nTestView(View):
    template_name = "i18n_test.jinja"

    def get(self, request, data=None):
        class Author:
            name = "Freddy Fred"

        class Book:
            title = "Big 'ol Book"

        return render(request, self.template_name, {
            "v_index": "Index",
            "table_sort": lambda x, y: f"{x} {y}",
            "invoice_count": 1,
            "trimmed_invoice_count": 2,
            "author": Author(),
            "book": Book(),
        })

class I18nTestViewDTL(I18nTestView):
    template_name = "i18n_test.html"

class ContextManipulationTestView(View):
    def get(self, request):
        return render(request, "hello_world.jinja", {"name": "Jinja2"})

# ==== generic.detail ====
class DetailTestView(DetailView):
    model = TestModel

# ==== generic.edit ====
class CreateTestView(CreateView):
    model = TestModel
    fields = []
    template_name_suffix = '_create'

class DeleteTestView(DeleteView):
    model = TestModel

class UpdateTestView(UpdateView):
    model = TestModel
    fields = []
    template_name_suffix = '_update'

# ==== generic.list ====
class ListTestView(ListView):
    model = TestModel

# ==== generic.dates ====
class ArchiveIndexTestView(ArchiveIndexView):
    model = TestModel
    date_field = 'date'

class YearArchiveTestView(YearArchiveView):
    model = TestModel
    date_field = 'date'

class MonthArchiveTestView(MonthArchiveView):
    model = TestModel
    date_field = 'date'

class WeekArchiveTestView(WeekArchiveView):
    model = TestModel
    date_field = 'date'

class DayArchiveTestView(DayArchiveView):
    model = TestModel
    date_field = 'date'

class TodayArchiveTestView(TodayArchiveView):
    model = TestModel
    date_field = 'date'
    template_name_suffix = '_archive_today'

class DateDetailTestView(DateDetailView):
    model = TestModel
    date_field = 'date'
    template_name_suffix = '_date_detail'

class StreamingTestView(View):
    def get(self, request, *args, **kwargs):
        context = {"name": "Streaming Jinja2", "view": type(self)}
        template = loader.get_template('streaming_test.jinja')
        return StreamingHttpResponse(template.stream(context, request), content_type='text/html')



from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='testmodel',
            name='date',
            field=models.DateTimeField(null=True),
            preserve_default=True,
        ),
    ]



from django.core.cache import caches
from django.utils.functional import cached_property
from jinja2 import BytecodeCache as _BytecodeCache


class BytecodeCache(_BytecodeCache):
    """
    A bytecode cache for Jinja2 that uses Django's caching framework.
    """

    def __init__(self, cache_name):
        self._cache_name = cache_name

    @cached_property
    def backend(self):
        return caches[self._cache_name]

    def load_bytecode(self, bucket):
        key = f'jinja2_{str(bucket.key)}'
        bytecode = self.backend.get(key)
        if bytecode:
            bucket.bytecode_from_string(bytecode)

    def dump_bytecode(self, bucket):
        key = f'jinja2_{str(bucket.key)}'
        self.backend.set(key, bucket.bytecode_to_string())



from django.views.generic.edit import CreateView as _django_CreateView
from django.views.generic.edit import DeleteView as _django_DeleteView
from django.views.generic.edit import UpdateView as _django_UpdateView

from .base import Jinja2TemplateResponseMixin


class CreateView(Jinja2TemplateResponseMixin, _django_CreateView):
    pass

class DeleteView(Jinja2TemplateResponseMixin, _django_DeleteView):
    pass

class UpdateView(Jinja2TemplateResponseMixin, _django_UpdateView):
    pass



from django_jinja import library
from jinja2 import pass_context
from subdomains.templatetags.subdomainurls import url as subdomain_url


@library.global_function
@pass_context
def url(context, *args, **kwargs):
    return subdomain_url(context, *args, **kwargs)



from .dates import (ArchiveIndexView, YearArchiveView, MonthArchiveView,
                    WeekArchiveView, DayArchiveView, TodayArchiveView,
                    DateDetailView)
from .detail import DetailView
from .edit import CreateView, UpdateView, DeleteView
from .list import ListView


__all__ = [
    'ArchiveIndexView', 'YearArchiveView', 'MonthArchiveView',
    'WeekArchiveView', 'DayArchiveView', 'TodayArchiveView',
    'DateDetailView', 'DetailView', 'CreateView', 'UpdateView',
    'DeleteView', 'ListView',
]



from django_jinja import library
import jinja2


@library.test(name="one")
def is_one(n):
    return n == 1


@library.filter
@jinja2.pass_context
def replace(context, value, x, y):
    return value.replace(x, y)


@library.global_function
def myecho(data):
    return data


@library.global_function
@library.render_with("test-render-with.jinja")
def myrenderwith(*args, **kwargs):
    return {"name": "Foo"}



#!/usr/bin/env python3

from setuptools import setup

setup(
    name = "django-jinja",
    version = "2.10.2",
    description = "Jinja2 templating language integrated in Django.",
    long_description = open("README.rst").read(),
    long_description_content_type='text/x-rst',
    keywords = "django, jinja2",
    author = "Andrey Antukh",
    author_email = "niwi@niwi.be",
    maintainer = "Asif Saif Uddin",
    maintainer_email = "auvipy@gmail.com",
    url = "https://github.com/niwinz/django-jinja",
    license = "BSD",
    packages = [
        "django_jinja",
        "django_jinja.builtins",
        "django_jinja.management",
        "django_jinja.management.commands",
        "django_jinja.contrib",
        "django_jinja.contrib._easy_thumbnails",
        "django_jinja.contrib._easy_thumbnails.templatetags",
        "django_jinja.contrib._humanize",
        "django_jinja.contrib._humanize.templatetags",
        "django_jinja.contrib._subdomains",
        "django_jinja.contrib._subdomains.templatetags",
        "django_jinja.views",
        "django_jinja.views.generic",
    ],
    include_package_data = True,
    python_requires = ">=3.6",
    install_requires = [
        "jinja2>=3",
        "django>=2.2",
    ],
    tests_require = [
        "pytz",
    ],
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Internet :: WWW/HTTP",
    ]
)



from django.contrib.humanize.templatetags import humanize
from django_jinja import library


@library.filter
def ordinal(source):
    return humanize.ordinal(source)


@library.filter
def intcomma(source, use_l10n=True):
    return humanize.intcomma(source, use_l10n)


@library.filter
def intword(source):
    return humanize.intword(source)


@library.filter
def apnumber(source):
    return humanize.apnumber(source)


@library.filter
def naturalday(source, arg=None):
    return humanize.naturalday(source, arg)


@library.filter
def naturaltime(source):
    return humanize.naturaltime(source)



#!/usr/bin/env python
import os, sys
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "settings")

if __name__ == "__main__":
    from django.core.management import execute_from_command_line
    args = sys.argv
    args.insert(1, "test")

    if len(args) == 2:
        args.insert(2, "testapp")

    execute_from_command_line(args)



import os, sys
sys.path.insert(0, "..")

BASE_DIR = os.path.dirname(__file__)
DEBUG = False
TEMPLATE_DEBUG = False

ALLOWED_HOSTS = ["*"]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "foobar.db"
    }
}
DEFAULT_AUTO_FIELD = "django.db.models.AutoField"

MIDDLEWARE_CLASSES = [
    # "django.middleware.csrf.CsrfViewMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]

ROOT_URLCONF = "testapp.urls"

USE_I18N =  True
USE_TZ = True
TIME_ZONE = "UTC"
LANGUAGE_CODE = "en"
ADMIN_MEDIA_PREFIX = "/static/admin/"
INTERNAL_IPS = ("127.0.0.1",)

STATIC_ROOT = os.path.join(BASE_DIR, "static")
STATIC_URL = "/static/"

SESSION_ENGINE = "django.contrib.sessions.backends.signed_cookies"
SESSION_EXPIRE_AT_BROWSER_CLOSE = True

STATICFILES_FINDERS = (
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
)

# TEMPLATE_DIRS = ()

SECRET_KEY = "di!n($kqa3)nd%ikad#kcjpkd^uw*h%*kj=*pm7$vbo6ir7h=l"
INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "django.contrib.messages",
    "django_jinja",
    "testapp",
)

from django_jinja.builtins import DEFAULT_EXTENSIONS
JINJA2_MUTE_URLRESOLVE_EXCEPTIONS = True

TEMPLATES = [
    {
        "BACKEND": "django_jinja.backend.Jinja2",
        "NAME": "jinja2",
        "APP_DIRS": True,
        "OPTIONS": {
            "debug": True,
            "context_processors": [
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
            ],
            "constants": {
                "foo": "bar",
            },
            "policies": {
                "ext.i18n.trimmed": True,
            },
            "extensions": DEFAULT_EXTENSIONS + [
                "django_jinja.builtins.extensions.DjangoExtraFiltersExtension",
            ]
        }
    },
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True
    }
]

TEST_RUNNER = "django.test.runner.DiscoverRunner"


LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "filters": {
        "require_debug_false": {
            "()": "django.utils.log.RequireDebugFalse"
        }
    },

    "formatters": {
        "simple": {
            "format": "%(asctime)s: %(message)s",
        }
    },

    "handlers": {
        "console":{
            "level":"DEBUG",
            "class":"logging.StreamHandler",
            "formatter": "simple"
        },

    },
    "loggers": {
        "django.request": {
            "handlers": ["console"],
            "level": "ERROR",
            "propagate": True,
        },
        "arandomtable.custom": {
            "handlers": ["console"],
            "level": "DEBUG",
        },

    }
}



from django.views.generic.dates import ArchiveIndexView as _django_ArchiveIndexView
from django.views.generic.dates import YearArchiveView as _django_YearArchiveView
from django.views.generic.dates import MonthArchiveView as _django_MonthArchiveView
from django.views.generic.dates import WeekArchiveView as _django_WeekArchiveView
from django.views.generic.dates import DayArchiveView as _django_DayArchiveView
from django.views.generic.dates import TodayArchiveView as _django_TodayArchiveView
from django.views.generic.dates import DateDetailView as _django_DateDetailView

from .base import Jinja2TemplateResponseMixin


class ArchiveIndexView(Jinja2TemplateResponseMixin, _django_ArchiveIndexView):
    pass

class YearArchiveView(Jinja2TemplateResponseMixin, _django_YearArchiveView):
    pass

class MonthArchiveView(Jinja2TemplateResponseMixin, _django_MonthArchiveView):
    pass

class WeekArchiveView(Jinja2TemplateResponseMixin, _django_WeekArchiveView):
    pass

class DayArchiveView(Jinja2TemplateResponseMixin, _django_DayArchiveView):
    pass

class TodayArchiveView(Jinja2TemplateResponseMixin, _django_TodayArchiveView):
    pass

class DateDetailView(Jinja2TemplateResponseMixin, _django_DateDetailView):
    pass



import logging

from easy_thumbnails.conf import settings
from easy_thumbnails.templatetags import thumbnail as _thumbnail
from django_jinja import library
from functools import wraps


logger = logging.getLogger(__name__)


def debug_silence(error_output=''):
    def inner(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except Exception as exc:
                if settings.THUMBNAIL_DEBUG:
                    raise
                logger.error('Error: %s', exc)
                return error_output
        return wrapper
    return inner


@library.filter
@debug_silence(error_output='')
def thumbnail_url(source, alias):
    return _thumbnail.thumbnail_url(source, alias)


@library.global_function
@debug_silence(error_output=None)
def thumbnailer_passive(obj):
    return _thumbnail.thumbnailer_passive(obj)


@library.global_function
@debug_silence(error_output=None)
def thumbnailer(obj):
    return _thumbnail.thumbnailer(obj)


@library.global_function
@debug_silence(error_output='')
def thumbnail(source, **kwargs):
    thumbnail =  _thumbnail.get_thumbnailer(source).get_thumbnail(kwargs)
    return thumbnail.url
